import { createPriorityQueue } from '@algorithm.ts/priority-queue';

var GomokuDirectionType;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(GomokuDirectionType || (GomokuDirectionType = {}));
const GomokuDirections = Array.from(Object.entries({
    [GomokuDirectionType.TOP]: [-1, 0],
    [GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [GomokuDirectionType.RIGHT]: [0, 1],
    [GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [GomokuDirectionType.BOTTOM]: [1, 0],
    [GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [GomokuDirectionType.LEFT]: [0, -1],
    [GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const GomokuDirectionTypes = {
    full: [
        GomokuDirectionType.LEFT,
        GomokuDirectionType.TOP_LEFT,
        GomokuDirectionType.TOP,
        GomokuDirectionType.TOP_RIGHT,
        GomokuDirectionType.RIGHT,
        GomokuDirectionType.BOTTOM_RIGHT,
        GomokuDirectionType.BOTTOM,
        GomokuDirectionType.BOTTOM_LEFT,
    ],
    leftHalf: [
        GomokuDirectionType.LEFT,
        GomokuDirectionType.TOP_LEFT,
        GomokuDirectionType.TOP,
        GomokuDirectionType.TOP_RIGHT,
    ],
    rightHalf: [
        GomokuDirectionType.RIGHT,
        GomokuDirectionType.BOTTOM_RIGHT,
        GomokuDirectionType.BOTTOM,
        GomokuDirectionType.BOTTOM_LEFT,
    ],
};
const GomokuDirectionTypeBitset = {
    full: GomokuDirectionTypes.full.reduce((acc, dirType) => acc | (1 << dirType), 0),
    leftHalf: GomokuDirectionTypes.leftHalf.reduce((acc, dirType) => acc | (1 << dirType), 0),
    rightHalf: GomokuDirectionTypes.rightHalf.reduce((acc, dirType) => acc | (1 << dirType), 0),
};

const { full: fullDirectionTypes$2 } = GomokuDirectionTypes;
class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_ADJACENT;
    MAX_DISTANCE_OF_NEIGHBOR;
    TOTAL_POS;
    MIDDLE_POS;
    board;
    idx;
    _idxMap;
    _gomokuDirections;
    _maxMovableMap;
    _dirStartPosMap;
    _dirStartPosSet;
    _dirNeighborSet;
    _neighborPlacedCount;
    _placedCount;
    constructor(props) {
        const { MAX_ROW, MAX_COL, MAX_ADJACENT, MAX_DISTANCE_OF_NEIGHBOR } = props;
        const _MAX_ROW = Math.max(1, MAX_ROW);
        const _MAX_COL = Math.max(1, MAX_COL);
        const _MAX_ADJACENT = Math.max(1, MAX_ADJACENT);
        const _MAX_DISTANCE_OF_NEIGHBOR = Math.max(1, MAX_DISTANCE_OF_NEIGHBOR);
        const _TOTAL_POS = _MAX_ROW * _MAX_COL;
        const idx = (r, c) => r * _MAX_ROW + c;
        this.MAX_ROW = _MAX_ROW;
        this.MAX_COL = _MAX_COL;
        this.MAX_ADJACENT = _MAX_ADJACENT;
        this.MAX_DISTANCE_OF_NEIGHBOR = _MAX_DISTANCE_OF_NEIGHBOR;
        this.TOTAL_POS = _TOTAL_POS;
        this.MIDDLE_POS = _TOTAL_POS >> 1;
        this.board = new Array(_TOTAL_POS).fill(-1);
        this.idx = idx;
        const _idxMap = new Array(_TOTAL_POS);
        for (let r = 0; r < _MAX_ROW; ++r) {
            for (let c = 0; c < _MAX_COL; ++c) {
                const posId = idx(r, c);
                _idxMap[posId] = [r, c];
            }
        }
        const _gomokuDirections = GomokuDirections.map(([dr, dc]) => dr * MAX_ROW + dc);
        const _maxMovableMap = new Array(fullDirectionTypes$2.length)
            .fill([])
            .map(() => new Array(_TOTAL_POS).fill(0));
        const _dirStartPosSet = new Array(fullDirectionTypes$2.length).fill([]).map(() => []);
        const _dirStartPosMap = new Array(fullDirectionTypes$2.length)
            .fill([])
            .map(() => new Array(_TOTAL_POS).fill(0));
        this.traverseAllDirections(dirType => {
            const revDirType = dirType ^ 1;
            const [dr, dc] = GomokuDirections[dirType];
            return posId => {
                const [r, c] = _idxMap[posId];
                const r2 = r + dr;
                const c2 = c + dc;
                if (r2 < 0 || r2 >= _MAX_ROW || c2 < 0 || c2 >= _MAX_ROW) {
                    _maxMovableMap[dirType][posId] = 0;
                    _dirStartPosMap[revDirType][posId] = posId;
                    _dirStartPosSet[revDirType].push(posId);
                }
                else {
                    const posId2 = idx(r2, c2);
                    _maxMovableMap[dirType][posId] = _maxMovableMap[dirType][posId2] + 1;
                    _dirStartPosMap[revDirType][posId] = _dirStartPosMap[revDirType][posId2];
                }
            };
        });
        const _neighborMap = new Array(_TOTAL_POS);
        for (let posId = 0; posId < _TOTAL_POS; ++posId) {
            const neighbors = [];
            _neighborMap[posId] = neighbors;
            for (const dirType of fullDirectionTypes$2) {
                let posId2 = posId;
                for (let step = 0; step < _MAX_DISTANCE_OF_NEIGHBOR; ++step) {
                    if (1 > _maxMovableMap[dirType][posId2])
                        break;
                    posId2 += _gomokuDirections[dirType];
                    neighbors.push(posId2);
                }
            }
        }
        this._gomokuDirections = _gomokuDirections;
        this._idxMap = _idxMap;
        this._maxMovableMap = _maxMovableMap;
        this._dirStartPosMap = _dirStartPosMap;
        this._dirStartPosSet = _dirStartPosSet;
        this._dirNeighborSet = _neighborMap;
        this._neighborPlacedCount = new Array(_TOTAL_POS).fill(0);
        this._placedCount = 0;
    }
    get placedCount() {
        return this._placedCount;
    }
    init(pieces) {
        const board = this.board;
        board.fill(-1);
        this._placedCount = 0;
        const { _neighborPlacedCount } = this;
        _neighborPlacedCount.fill(0);
        for (const { r, c, p } of pieces) {
            const posId = this.idx(r, c);
            if (board[posId] < 0) {
                board[posId] = p;
                this._placedCount += 1;
                for (const neighborId of this.accessibleNeighbors(posId)) {
                    _neighborPlacedCount[neighborId] += 1;
                }
            }
        }
    }
    forward(posId, playerId) {
        const board = this.board;
        if (posId < 0 || posId >= this.TOTAL_POS || board[posId] >= 0)
            return false;
        board[posId] = playerId;
        this._placedCount += 1;
        for (const neighborId of this.accessibleNeighbors(posId)) {
            this._neighborPlacedCount[neighborId] += 1;
        }
        return true;
    }
    revert(posId) {
        const board = this.board;
        if (posId < 0 || posId >= this.TOTAL_POS || board[posId] < 0)
            return false;
        board[posId] = -1;
        this._placedCount -= 1;
        for (const neighborId of this.accessibleNeighbors(posId)) {
            this._neighborPlacedCount[neighborId] -= 1;
        }
        return true;
    }
    revIdx(posId) {
        return this._idxMap[posId];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    safeMove(posId, dirType, step) {
        return step <= this._maxMovableMap[dirType][posId]
            ? posId + this._gomokuDirections[dirType] * step
            : -1;
    }
    safeMoveOneStep(posId, dirType) {
        return 1 <= this._maxMovableMap[dirType][posId] ? posId + this._gomokuDirections[dirType] : -1;
    }
    fastMove(posId, dirType, step) {
        return posId + this._gomokuDirections[dirType] * step;
    }
    fastMoveOneStep(posId, dirType) {
        return posId + this._gomokuDirections[dirType];
    }
    maxMovableSteps(posId, dirType) {
        return this._maxMovableMap[dirType][posId];
    }
    accessibleNeighbors(posId) {
        return this._dirNeighborSet[posId];
    }
    hasPlacedNeighbors(posId) {
        return this._neighborPlacedCount[posId] > 0;
    }
    getStartPosId(posId, dirType) {
        return this._dirStartPosMap[dirType][posId];
    }
    getStartPosSet(dirType) {
        return this._dirStartPosSet[dirType];
    }
    traverseAllDirections(handle) {
        const { TOTAL_POS } = this;
        const { leftHalf, rightHalf } = GomokuDirectionTypes;
        for (const dirType of leftHalf) {
            const h = handle(dirType);
            for (let posId = 0; posId < TOTAL_POS; ++posId)
                h(posId);
        }
        for (const dirType of rightHalf) {
            const h = handle(dirType);
            for (let posId = TOTAL_POS - 1; posId >= 0; --posId)
                h(posId);
        }
    }
}

const { full: fullDirectionTypes$1, rightHalf: halfDirectionTypes$1 } = GomokuDirectionTypes;
const { rightHalf: allDirectionTypeBitset$1 } = GomokuDirectionTypeBitset;
class GomokuCountMap {
    context;
    _mustDropPosSet;
    _candidateCouldReachFinal;
    _rightHalfDirCountMap;
    constructor(context) {
        this.context = context;
        this._mustDropPosSet = new Array(2).fill([]).map(() => new Set());
        this._candidateCouldReachFinal = new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]);
        this._rightHalfDirCountMap = new Array(fullDirectionTypes$1.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => []));
    }
    init() {
        const { context, _mustDropPosSet, _candidateCouldReachFinal } = this;
        const { TOTAL_POS, board } = context;
        const { _rightHalfDirCountMap } = this;
        for (const dirType of halfDirectionTypes$1) {
            for (const startPosId of context.getStartPosSet(dirType)) {
                const counters = _rightHalfDirCountMap[dirType][startPosId];
                let index = 0;
                const maxSteps = context.maxMovableSteps(startPosId, dirType) + 1;
                for (let i = 0, posId = startPosId, i2, posId2; i < maxSteps; i = i2, posId = posId2) {
                    const playerId = board[posId];
                    for (i2 = i + 1, posId2 = posId; i2 < maxSteps; ++i2) {
                        posId2 = context.fastMoveOneStep(posId2, dirType);
                        if (board[posId2] !== playerId)
                            break;
                    }
                    counters[index++] = { playerId, count: i2 - i };
                }
                counters.length = index;
            }
        }
        _mustDropPosSet.forEach(set => set.clear());
        _candidateCouldReachFinal.forEach(item => item.fill(0));
        for (let posId = 0; posId < TOTAL_POS; ++posId) {
            if (context.board[posId] >= 0)
                continue;
            let data0 = 0;
            let data1 = 0;
            for (const dirType of halfDirectionTypes$1) {
                const bitFlag = 1 << dirType;
                if (this._couldReachFinalInDirection(0, posId, dirType))
                    data0 |= bitFlag;
                if (this._couldReachFinalInDirection(1, posId, dirType))
                    data1 |= bitFlag;
            }
            if (data0 > 0)
                _mustDropPosSet[0].add(posId);
            if (data1 > 0)
                _mustDropPosSet[1].add(posId);
            _candidateCouldReachFinal[posId][0] = data0;
            _candidateCouldReachFinal[posId][1] = data1;
        }
    }
    forward(posId) {
        const playerId = this.context.board[posId];
        for (const dirType of halfDirectionTypes$1) {
            this._updateHalfDirCounter(playerId, posId, dirType);
        }
        this._updateRelatedCouldReachFinal(posId);
    }
    revert(posId) {
        for (const dirType of halfDirectionTypes$1) {
            this._updateHalfDirCounter(-1, posId, dirType);
        }
        this._updateRelatedCouldReachFinal(posId);
    }
    getDirCounters(startPosId, dirType) {
        return this._rightHalfDirCountMap[dirType][startPosId];
    }
    mustDropPos(playerId) {
        return this._mustDropPosSet[playerId];
    }
    candidateCouldReachFinal(playerId, posId) {
        return this._candidateCouldReachFinal[posId][playerId] > 0;
    }
    _updateRelatedCouldReachFinal(centerPosId) {
        const { context } = this;
        const { board } = context;
        this._updateCouldReachFinal(centerPosId, allDirectionTypeBitset$1);
        for (const dirType of halfDirectionTypes$1) {
            const expiredBitset = 1 << dirType;
            const revDirType = dirType ^ 1;
            const maxMovableSteps0 = context.maxMovableSteps(centerPosId, revDirType);
            for (let posId = centerPosId, step = 0; step < maxMovableSteps0; ++step) {
                posId = context.fastMoveOneStep(posId, revDirType);
                if (board[posId] < 0) {
                    this._updateCouldReachFinal(posId, expiredBitset);
                    break;
                }
            }
            const maxMovableSteps2 = context.maxMovableSteps(centerPosId, dirType);
            for (let posId = centerPosId, step = 0; step < maxMovableSteps2; ++step) {
                posId = context.fastMoveOneStep(posId, dirType);
                if (board[posId] < 0) {
                    this._updateCouldReachFinal(posId, expiredBitset);
                    break;
                }
            }
        }
    }
    _updateCouldReachFinal(posId, expiredBitset) {
        if (expiredBitset > 0) {
            const { _mustDropPosSet, _candidateCouldReachFinal } = this;
            const prevData0 = _candidateCouldReachFinal[posId][0];
            const prevData1 = _candidateCouldReachFinal[posId][1];
            _mustDropPosSet[0].delete(posId);
            _mustDropPosSet[1].delete(posId);
            if (this.context.board[posId] >= 0) {
                _candidateCouldReachFinal[posId][0] = 0;
                _candidateCouldReachFinal[posId][1] = 0;
                return;
            }
            let nextData0 = 0;
            let nextData1 = 0;
            for (const dirType of halfDirectionTypes$1) {
                const bitFlag = 1 << dirType;
                if (bitFlag & expiredBitset) {
                    if (this._couldReachFinalInDirection(0, posId, dirType))
                        nextData0 |= bitFlag;
                    if (this._couldReachFinalInDirection(1, posId, dirType))
                        nextData1 |= bitFlag;
                }
                else {
                    nextData0 |= prevData0 & bitFlag;
                    nextData1 |= prevData1 & bitFlag;
                }
            }
            _candidateCouldReachFinal[posId][0] = nextData0;
            _candidateCouldReachFinal[posId][1] = nextData1;
            if (nextData0 > 0)
                _mustDropPosSet[0].add(posId);
            if (nextData1 > 0)
                _mustDropPosSet[1].add(posId);
        }
    }
    _couldReachFinalInDirection(playerId, posId, dirType) {
        const { context } = this;
        const { MAX_ADJACENT, board } = context;
        const revDirType = dirType ^ 1;
        let count = 1;
        const maxMovableSteps0 = context.maxMovableSteps(posId, revDirType);
        for (let id = posId, step = 0; step < maxMovableSteps0; ++step) {
            id = context.fastMoveOneStep(id, revDirType);
            if (board[id] !== playerId)
                break;
            count += 1;
        }
        const maxMovableSteps2 = context.maxMovableSteps(posId, dirType);
        for (let id = posId, step = 0; step < maxMovableSteps2; ++step) {
            id = context.fastMoveOneStep(id, dirType);
            if (board[id] !== playerId)
                break;
            count += 1;
        }
        return count >= MAX_ADJACENT;
    }
    _updateHalfDirCounter(playerId, posId, dirType) {
        const { context } = this;
        const revDirType = dirType ^ 1;
        const startPosId = context.getStartPosId(posId, dirType);
        const counters = this._rightHalfDirCountMap[dirType][startPosId];
        let index = 0;
        let remain = context.maxMovableSteps(posId, revDirType) + 1;
        for (; index < counters.length; ++index) {
            const cnt = counters[index].count;
            if (cnt >= remain)
                break;
            remain -= cnt;
        }
        if (remain === 1) {
            if (counters[index].count === 1) {
                const isSameWithLeft = index > 0 && counters[index - 1].playerId === playerId;
                const isSameWithRight = index + 1 < counters.length && counters[index + 1].playerId === playerId;
                if (isSameWithLeft) {
                    if (isSameWithRight) {
                        counters[index - 1].count += 1 + counters[index + 1].count;
                        counters.splice(index, 2);
                    }
                    else {
                        counters[index - 1].count += 1;
                        counters.splice(index, 1);
                    }
                }
                else {
                    if (isSameWithRight) {
                        counters[index + 1].count += 1;
                        counters.splice(index, 1);
                    }
                    else {
                        counters[index].playerId = playerId;
                    }
                }
            }
            else {
                counters[index].count -= 1;
                if (index > 0 && counters[index - 1].playerId === playerId) {
                    counters[index - 1].count += 1;
                }
                else {
                    counters.splice(index, 0, { playerId, count: 1 });
                }
            }
        }
        else if (remain < counters[index].count) {
            const { playerId: hitPlayerId, count: hitCount } = counters[index];
            counters.splice(index, 1, { playerId: hitPlayerId, count: remain - 1 }, { playerId, count: 1 }, { playerId: hitPlayerId, count: hitCount - remain });
        }
        else {
            counters[index].count -= 1;
            if (index + 1 < counters.length && counters[index + 1].playerId === playerId) {
                counters[index + 1].count += 1;
            }
            else {
                counters.splice(index + 1, 0, { playerId, count: 1 });
            }
        }
    }
}

const { full: fullDirectionTypes, rightHalf: halfDirectionTypes } = GomokuDirectionTypes;
const { rightHalf: allDirectionTypeBitset } = GomokuDirectionTypeBitset;
class GomokuState {
    context;
    NEXT_MOVER_BUFFER = 2;
    _countMap;
    _scoreMap;
    _countOfReachLimits;
    _countOfReachLimitsDirMap;
    _stateScoreMap;
    _stateScoreDirMap;
    _candidateSet;
    _candidateQueues;
    _candidateLatestIdMap;
    _candidateScoreDirMap;
    _candidateScoreExpired;
    _candidateUuid;
    constructor(props) {
        const { context, scoreMap } = props;
        const _candidateQueues = new Array(2)
            .fill([])
            .map(() => createPriorityQueue((x, y) => x.score - y.score));
        _candidateQueues.forEach(Q => Q.init());
        const _candidateLatestIdMap = new Array(context.TOTAL_POS).fill(0);
        const _countOfReachLimitsDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _stateScoreDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _candidateScoreDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _candidateScoreExpired = new Array(context.TOTAL_POS).fill(allDirectionTypeBitset);
        this.context = context;
        this._candidateUuid = 0;
        this._countMap = new GomokuCountMap(context);
        this._scoreMap = scoreMap;
        this._countOfReachLimits = [0, 0];
        this._countOfReachLimitsDirMap = _countOfReachLimitsDirMap;
        this._stateScoreMap = [0, 0];
        this._stateScoreDirMap = _stateScoreDirMap;
        this._candidateQueues = _candidateQueues;
        this._candidateLatestIdMap = _candidateLatestIdMap;
        this._candidateSet = new Set();
        this._candidateScoreDirMap = _candidateScoreDirMap;
        this._candidateScoreExpired = _candidateScoreExpired;
    }
    init(pieces) {
        const { context, _countOfReachLimits, _countOfReachLimitsDirMap, _stateScoreMap, _stateScoreDirMap, _candidateLatestIdMap, _candidateQueues, _candidateSet, _candidateScoreExpired, } = this;
        this._candidateUuid = 0;
        this._countMap.init();
        _candidateLatestIdMap.fill(0);
        _candidateQueues.forEach(Q => Q.init());
        _countOfReachLimits.fill(0);
        _stateScoreMap.fill(0);
        for (const dirType of halfDirectionTypes) {
            for (const startPosId of context.getStartPosSet(dirType)) {
                const { scores, countOfReachLimits } = this._evaluateScoreInDirection(startPosId, dirType);
                _countOfReachLimits[0] += countOfReachLimits[0];
                _countOfReachLimits[1] += countOfReachLimits[1];
                _countOfReachLimitsDirMap[dirType][startPosId][0] = countOfReachLimits[0];
                _countOfReachLimitsDirMap[dirType][startPosId][1] = countOfReachLimits[1];
                _stateScoreMap[0] += scores[0];
                _stateScoreMap[1] += scores[1];
                _stateScoreDirMap[dirType][startPosId][0] = scores[0];
                _stateScoreDirMap[dirType][startPosId][1] = scores[1];
            }
        }
        _candidateSet.clear();
        for (const { r, c } of pieces) {
            const posId = context.idx(r, c);
            for (const posId2 of context.accessibleNeighbors(posId)) {
                if (context.board[posId2] < 0)
                    _candidateSet.add(posId2);
            }
        }
        for (let posId = 0; posId < context.TOTAL_POS; ++posId) {
            _candidateScoreExpired[posId] = allDirectionTypeBitset;
        }
    }
    forward(posId) {
        this._countMap.forward(posId);
        const { context, _candidateSet } = this;
        _candidateSet.delete(posId);
        for (const posId2 of context.accessibleNeighbors(posId)) {
            if (context.board[posId2] < 0)
                _candidateSet.add(posId2);
        }
        this._updateStateScore(posId);
        this._expireCandidates(posId);
    }
    revert(posId) {
        this._countMap.revert(posId);
        const { context, _candidateSet } = this;
        if (context.hasPlacedNeighbors(posId))
            _candidateSet.add(posId);
        for (const posId2 of context.accessibleNeighbors(posId)) {
            if (!context.hasPlacedNeighbors(posId2))
                _candidateSet.delete(posId2);
        }
        this._updateStateScore(posId);
        this._expireCandidates(posId);
    }
    expand(nextPlayerId, candidates, MAX_SIZE) {
        const { context, _candidateQueues, _candidateLatestIdMap, _candidateSet } = this;
        const topCandidate = this.topCandidate(nextPlayerId);
        if (topCandidate !== undefined && topCandidate.score === Number.MAX_VALUE) {
            candidates[0] = topCandidate;
            return 1;
        }
        if (context.board[context.MIDDLE_POS] < 0)
            _candidateSet.add(this.context.MIDDLE_POS);
        for (const posId of _candidateSet) {
            this._reEvaluateCandidate(posId);
        }
        const _size = Math.min(_candidateSet.size, MAX_SIZE ?? _candidateSet.size);
        const Q = _candidateQueues[nextPlayerId];
        for (let i = 0, item; i < _size; ++i) {
            for (item = Q.dequeue();; item = Q.dequeue()) {
                if (_candidateSet.has(item.posId) && _candidateLatestIdMap[item.posId] === item.$id)
                    break;
            }
            candidates[i] = item;
        }
        Q.init(candidates, 0, _size);
        return _size;
    }
    topCandidate(nextPlayerId) {
        if (this._candidateSet.size <= 0)
            return undefined;
        const { _candidateSet, _countMap } = this;
        for (const posId of _countMap.mustDropPos(nextPlayerId)) {
            return { $id: -1, posId, score: Number.MAX_VALUE };
        }
        for (const posId of _countMap.mustDropPos(nextPlayerId ^ 1)) {
            return { $id: -1, posId, score: Number.MAX_VALUE };
        }
        const { context, _candidateQueues, _candidateLatestIdMap } = this;
        if (context.board[context.MIDDLE_POS] < 0)
            _candidateSet.add(this.context.MIDDLE_POS);
        for (const posId of _candidateSet) {
            this._reEvaluateCandidate(posId);
        }
        const Q = _candidateQueues[nextPlayerId];
        for (let item;; Q.dequeue()) {
            item = Q.top();
            if (_candidateSet.has(item.posId) && _candidateLatestIdMap[item.posId] === item.$id) {
                return item;
            }
        }
    }
    score(currentPlayer, scoreForPlayer) {
        const score0 = this._stateScoreMap[scoreForPlayer];
        const score1 = this._stateScoreMap[scoreForPlayer ^ 1];
        const nextMoverFac = this.NEXT_MOVER_BUFFER;
        return currentPlayer === scoreForPlayer
            ? score0 - score1 * nextMoverFac
            : score0 * nextMoverFac - score1;
    }
    isWin(currentPlayer) {
        return this._countOfReachLimits[currentPlayer] > 0;
    }
    isDraw() {
        return this.context.placedCount === this.context.TOTAL_POS;
    }
    isFinal() {
        const { context, _countOfReachLimits } = this;
        if (context.placedCount === context.TOTAL_POS)
            return true;
        if (_countOfReachLimits[0] > 0 || _countOfReachLimits[1] > 0)
            return true;
        return false;
    }
    _updateStateScore(posId) {
        const { context, _countOfReachLimits, _countOfReachLimitsDirMap, _stateScoreMap, _stateScoreDirMap, } = this;
        for (const dirType of halfDirectionTypes) {
            const startPosId = context.getStartPosId(posId, dirType);
            const { scores: [score0, score1], countOfReachLimits: [count0, count1], } = this._evaluateScoreInDirection(startPosId, dirType);
            _countOfReachLimits[0] += count0 - _countOfReachLimitsDirMap[dirType][startPosId][0];
            _countOfReachLimits[1] += count1 - _countOfReachLimitsDirMap[dirType][startPosId][1];
            _countOfReachLimitsDirMap[dirType][startPosId][0] = count0;
            _countOfReachLimitsDirMap[dirType][startPosId][1] = count1;
            _stateScoreMap[0] += score0 - _stateScoreDirMap[dirType][startPosId][0];
            _stateScoreMap[1] += score1 - _stateScoreDirMap[dirType][startPosId][1];
            _stateScoreDirMap[dirType][startPosId][0] = score0;
            _stateScoreDirMap[dirType][startPosId][1] = score1;
        }
    }
    _expireCandidates(posId) {
        const { context, _candidateScoreExpired } = this;
        for (const dirType of halfDirectionTypes) {
            const startPosId = context.getStartPosId(posId, dirType);
            const maxSteps = context.maxMovableSteps(startPosId, dirType);
            const bitMark = 1 << dirType;
            for (let id = startPosId, step = 0;; ++step) {
                _candidateScoreExpired[id] |= bitMark;
                if (step === maxSteps)
                    break;
                id = context.fastMoveOneStep(id, dirType);
            }
        }
    }
    _reEvaluateCandidate(posId) {
        const { _candidateScoreExpired } = this;
        const expiredBitset = _candidateScoreExpired[posId];
        if (expiredBitset > 0) {
            const { context, _candidateQueues, _candidateLatestIdMap, _candidateScoreDirMap, _stateScoreDirMap, } = this;
            let prevScore0 = 0;
            let prevScore1 = 0;
            for (const dirType of halfDirectionTypes) {
                const startPosId = context.getStartPosId(posId, dirType);
                prevScore0 += _stateScoreDirMap[dirType][startPosId][0];
                prevScore1 += _stateScoreDirMap[dirType][startPosId][1];
            }
            let score0 = 0;
            this._forward(posId, 0);
            for (const dirType of halfDirectionTypes) {
                if ((1 << dirType) & expiredBitset) {
                    const startPosId = context.getStartPosId(posId, dirType);
                    const { scores } = this._evaluateScoreInDirection(startPosId, dirType);
                    _candidateScoreDirMap[dirType][posId][0] = scores[0];
                    score0 += scores[0];
                }
                else {
                    score0 += _candidateScoreDirMap[dirType][posId][0];
                }
            }
            this._revert(posId);
            let score1 = 0;
            this._forward(posId, 1);
            for (const dirType of halfDirectionTypes) {
                if ((1 << dirType) & expiredBitset) {
                    const startPosId = context.getStartPosId(posId, dirType);
                    const { scores } = this._evaluateScoreInDirection(startPosId, dirType);
                    _candidateScoreDirMap[dirType][posId][1] = scores[1];
                    score1 += scores[1];
                }
                else {
                    score1 += _candidateScoreDirMap[dirType][posId][1];
                }
            }
            this._revert(posId);
            const deltaScore0 = score0 - prevScore0;
            const deltaScore1 = score1 - prevScore1;
            const baseScore = deltaScore0 + deltaScore1;
            const candidateScore0 = baseScore + deltaScore0;
            const candidateScore1 = baseScore + deltaScore1;
            const uuid = ++this._candidateUuid;
            _candidateLatestIdMap[posId] = uuid;
            _candidateQueues[0].enqueue({ $id: uuid, posId, score: candidateScore0 });
            _candidateQueues[1].enqueue({ $id: uuid, posId, score: candidateScore1 });
            _candidateScoreExpired[posId] = 0;
        }
    }
    _evaluateScoreInDirection(startPosId, dirType) {
        const { context } = this;
        const { con, gap } = this._scoreMap;
        const { MAX_ADJACENT } = context;
        const THRESHOLD = MAX_ADJACENT - 1;
        const counters = this._countMap.getDirCounters(startPosId, dirType);
        const _size = counters.length;
        const scores = [0, 0];
        const countOfReachLimits = [0, 0];
        for (let i = 0, j, k; i < _size; i = j) {
            if (i > 0 && counters[i - 1].playerId < 0)
                i -= 1;
            k = i;
            j = i;
            let { playerId, count: maxPossibleCnt } = counters[j];
            if (playerId < 0) {
                if (++j === _size)
                    break;
                const { playerId: playerId2, count: cnt } = counters[j];
                playerId = playerId2;
                maxPossibleCnt += cnt;
                k += 1;
            }
            for (j += 1; j < _size; ++j) {
                const { playerId: playerId2, count: cnt } = counters[j];
                if (playerId2 >= 0 && playerId2 !== playerId)
                    break;
                maxPossibleCnt += cnt;
            }
            if (maxPossibleCnt >= MAX_ADJACENT) {
                let isFreeSide0 = k > i ? 1 : 0;
                for (let usedK = -1; k < j; k += 2, isFreeSide0 = 1) {
                    const cnt1 = counters[k].count;
                    if (cnt1 < THRESHOLD && k + 2 < j && counters[k + 1].count === 1) {
                        const cnt2 = counters[k + 2].count;
                        if (cnt2 < THRESHOLD) {
                            const isFreeSide2 = k + 3 < j ? 1 : 0;
                            const normalizedCnt = Math.min(cnt1 + cnt2, THRESHOLD);
                            scores[playerId] += gap[normalizedCnt][isFreeSide0 + isFreeSide2];
                            usedK = k + 2;
                            continue;
                        }
                    }
                    if (usedK !== k) {
                        let normalizedCnt = cnt1;
                        if (cnt1 >= MAX_ADJACENT) {
                            normalizedCnt = MAX_ADJACENT;
                            countOfReachLimits[playerId] += 1;
                        }
                        const isFreeSide2 = k + 1 < j ? 1 : 0;
                        scores[playerId] += con[normalizedCnt][isFreeSide0 + isFreeSide2];
                    }
                }
            }
        }
        return { scores, countOfReachLimits };
    }
    _forward(posId, playerId) {
        this.context.forward(posId, playerId);
        this._countMap.forward(posId);
    }
    _revert(posId) {
        this.context.revert(posId);
        this._countMap.revert(posId);
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

class GomokuStateCache {
    MAX_STATES;
    stateCompressor;
    states;
    constructor(TOTAL_POS, MAX_STATES = 2e5) {
        this.stateCompressor = new GomokuStateCompressor(TOTAL_POS);
        this.states = new Map();
        this.MAX_STATES = MAX_STATES;
    }
    get INITIAL_STATE() {
        return this.stateCompressor.INITIAL_STATE;
    }
    clear() {
        this.states.clear();
    }
    get(state) {
        return this.states.get(state);
    }
    set(state, score) {
        if (this.states.size < this.MAX_STATES)
            this.states.set(state, score);
    }
    calcNextState(cur, prevState, nextMovePosId) {
        const nextState = this.stateCompressor.compress(cur, prevState, BigInt(nextMovePosId));
        return nextState;
    }
}

const createScoreMap = (MAX_ADJACENT) => {
    const con = new Array(MAX_ADJACENT + 1).fill([]).map(() => [0, 0, 0]);
    const gap = new Array(MAX_ADJACENT + 1).fill([]).map(() => [0, 0, 0]);
    let BASE_VALUE = 2;
    for (let cnt = 1; cnt < MAX_ADJACENT; ++cnt, BASE_VALUE *= 16) {
        con[cnt] = [0, BASE_VALUE, BASE_VALUE * 2];
        con[cnt] = [0, BASE_VALUE / 2, BASE_VALUE];
    }
    const _v = con[MAX_ADJACENT - 1][1];
    gap[MAX_ADJACENT - 1] = [_v, _v, _v];
    con[MAX_ADJACENT] = [_v, _v, _v];
    con[MAX_ADJACENT] = [BASE_VALUE, BASE_VALUE, BASE_VALUE];
    return { con, gap };
};

class GomokuSolution {
    MAX_DEPTH_WIDE;
    MAX_DEPTH_NARROW;
    MAX_DEPTH_TIGHT;
    MAX_DEPTH_DEEP;
    POSSIBILITY_SEARCH_EQUIV_CANDIDATE;
    context;
    state;
    scoreMap;
    _CANDIDATE_SCORE_WIDE_MIN;
    _CANDIDATE_SCORE_NARROW_MIN;
    _CANDIDATE_SCORE_TIGHT_MIN;
    _CANDIDATE_SCORE_DEEP_MIN;
    _CANDIDATE_SCORE_DEEP_MAX;
    _cache;
    _candidatesList = [];
    _bestMoveId;
    _mainPlayerId;
    constructor(props) {
        const { MAX_ROW, MAX_COL, MAX_ADJACENT = 5, MAX_DEPTH_WIDE = 5, MAX_DEPTH_NARROW = 7, MAX_DEPTH_TIGHT = 9, MAX_DEPTH_DEEP = 24, MAX_DISTANCE_OF_NEIGHBOR = 2, POSSIBILITY_SEARCH_EQUIV_CANDIDATE = 0.98, } = props;
        const context = new GomokuContext({ MAX_ROW, MAX_COL, MAX_ADJACENT, MAX_DISTANCE_OF_NEIGHBOR });
        const scoreMap = props.scoreMap ?? createScoreMap(context.MAX_ADJACENT);
        const state = new GomokuState({ context, scoreMap });
        const _MAX_DEPTH_WIDE = Math.max(1, Math.round(MAX_DEPTH_WIDE));
        const _MAX_DEPTH_NARROW = Math.max(_MAX_DEPTH_WIDE, Math.round(MAX_DEPTH_NARROW));
        const _MAX_DEPTH_TIGHT = Math.max(_MAX_DEPTH_NARROW, Math.round(MAX_DEPTH_TIGHT));
        const _MAX_DEPTH_DEEP = Math.max(_MAX_DEPTH_TIGHT, Math.round(MAX_DEPTH_DEEP));
        const _POSSIBILITY_SEARCH_EQUIV_CANDIDATE = Math.min(1, Math.max(0, POSSIBILITY_SEARCH_EQUIV_CANDIDATE));
        const _candidatesList = new Array(context.TOTAL_POS + 1)
            .fill([])
            .map(() => []);
        this.MAX_DEPTH_WIDE = _MAX_DEPTH_WIDE;
        this.MAX_DEPTH_NARROW = _MAX_DEPTH_NARROW;
        this.MAX_DEPTH_TIGHT = _MAX_DEPTH_TIGHT;
        this.MAX_DEPTH_DEEP = _MAX_DEPTH_DEEP;
        this.POSSIBILITY_SEARCH_EQUIV_CANDIDATE = _POSSIBILITY_SEARCH_EQUIV_CANDIDATE;
        this.context = context;
        this.state = state;
        this.scoreMap = scoreMap;
        this._CANDIDATE_SCORE_WIDE_MIN = scoreMap.con[2][2] * 2;
        this._CANDIDATE_SCORE_NARROW_MIN = scoreMap.con[MAX_ADJACENT - 2][1];
        this._CANDIDATE_SCORE_TIGHT_MIN = scoreMap.con[MAX_ADJACENT - 2][2] * 2;
        this._CANDIDATE_SCORE_DEEP_MIN = scoreMap.con[MAX_ADJACENT - 1][1];
        this._CANDIDATE_SCORE_DEEP_MAX = scoreMap.con[MAX_ADJACENT][2] * 16;
        this._cache = new GomokuStateCache(BigInt(context.TOTAL_POS));
        this._candidatesList = _candidatesList;
        this._mainPlayerId = -1;
        this._bestMoveId = -1;
    }
    init(pieces) {
        this.context.init(pieces);
        this.state.init(pieces);
        this._cache.clear();
    }
    forward(r, c, playerId) {
        if (this.context.isValidPos(r, c)) {
            const posId = this.context.idx(r, c);
            this._forward(posId, playerId);
        }
    }
    revert(r, c) {
        if (this.context.isValidPos(r, c)) {
            const posId = this.context.idx(r, c);
            this._revert(posId);
        }
    }
    minimaxSearch(nextPlayer) {
        if (this.state.isFinal())
            return [-1, -1];
        this._cache.clear();
        this._bestMoveId = -1;
        this._mainPlayerId = nextPlayer;
        this._alphaBeta(nextPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, this._cache.INITIAL_STATE);
        if (this._bestMoveId < 0) {
            throw new Error('Oops! Something must be wrong, cannot find a valid moving strategy');
        }
        const [r, c] = this.context.revIdx(this._bestMoveId);
        return [r, c];
    }
    _alphaBeta(playerId, alpha, beta, cur, prevState) {
        return this._searchWideSpace(playerId, alpha, beta, cur, prevState);
    }
    _searchWideSpace(playerId, alpha, beta, cur, prevState) {
        const { _CANDIDATE_SCORE_WIDE_MIN, state, _cache, _mainPlayerId } = this;
        if (cur >= this.MAX_DEPTH_WIDE)
            return this._searchNarrowSpace(playerId, alpha, beta, cur, prevState);
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (state.isDraw())
            return Number.MAX_VALUE;
        const candidates = this._candidatesList[cur];
        const _size = state.expand(playerId, candidates, 16);
        if (cur === 0)
            this._bestMoveId = candidates[0].posId;
        const shouldCache = cur > 1;
        for (const candidate of this._selectOneCandidate(candidates, _size)) {
            const shouldSearchDeeper = candidate.score >= _CANDIDATE_SCORE_WIDE_MIN;
            const posId = candidate.posId;
            const nextState = _cache.calcNextState(cur, prevState, posId);
            let gamma = _cache.get(nextState);
            if (gamma === undefined) {
                this._forward(posId, playerId);
                gamma = shouldSearchDeeper
                    ? this._alphaBeta(playerId ^ 1, alpha, beta, cur + 1, nextState)
                    : this._score(playerId ^ 1);
                this._revert(posId);
                if (shouldCache)
                    _cache.set(nextState, gamma);
            }
            if (playerId === _mainPlayerId) {
                if (alpha < gamma) {
                    alpha = gamma;
                    if (cur === 0)
                        this._bestMoveId = posId;
                }
            }
            else {
                if (beta > gamma) {
                    beta = gamma;
                }
            }
            if (beta <= alpha)
                break;
        }
        return playerId === _mainPlayerId ? alpha : beta;
    }
    _searchNarrowSpace(playerId, alpha, beta, cur, prevState) {
        const { _CANDIDATE_SCORE_WIDE_MIN, state, _mainPlayerId, _cache } = this;
        if (cur >= this.MAX_DEPTH_NARROW)
            return this._searchTightSpace(playerId, alpha, beta, cur);
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (state.isDraw())
            return Number.MAX_VALUE;
        const candidates = this._candidatesList[cur];
        const _size = state.expand(playerId, candidates, 8);
        for (let i = 0; i < _size; ++i) {
            const candidate = candidates[i];
            const shouldSearchDeeper = candidate.score >= _CANDIDATE_SCORE_WIDE_MIN;
            const posId = candidate.posId;
            const nextState = _cache.calcNextState(cur, prevState, posId);
            let gamma = _cache.get(nextState);
            if (gamma === undefined) {
                this._forward(posId, playerId);
                gamma = shouldSearchDeeper
                    ? this._alphaBeta(playerId ^ 1, alpha, beta, cur + 1, nextState)
                    : this._score(playerId ^ 1);
                this._revert(posId);
                _cache.set(nextState, gamma);
            }
            if (playerId === _mainPlayerId) {
                if (alpha < gamma)
                    alpha = gamma;
            }
            else {
                if (beta > gamma)
                    beta = gamma;
            }
            if (beta <= alpha)
                break;
        }
        return playerId === _mainPlayerId ? alpha : beta;
    }
    _searchTightSpace(playerId, alpha, beta, cur) {
        const { _CANDIDATE_SCORE_TIGHT_MIN, state, _mainPlayerId } = this;
        if (cur >= this.MAX_DEPTH_TIGHT)
            return this._searchDeepSpace(playerId, cur);
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (state.isDraw())
            return Number.MAX_VALUE;
        const candidates = this._candidatesList[cur];
        const _size = state.expand(playerId, candidates, 2);
        for (let i = 0; i < _size; ++i) {
            const candidate = candidates[i];
            const shouldSearchDeeper = candidate.score >= _CANDIDATE_SCORE_TIGHT_MIN;
            const posId = candidate.posId;
            this._forward(posId, playerId);
            const gamma = shouldSearchDeeper
                ? this._searchTightSpace(playerId ^ 1, alpha, beta, cur + 1)
                : this._score(playerId ^ 1);
            this._revert(posId);
            if (playerId === _mainPlayerId) {
                if (alpha < gamma)
                    alpha = gamma;
            }
            else {
                if (beta > gamma)
                    beta = gamma;
            }
            if (beta <= alpha)
                break;
        }
        return playerId === _mainPlayerId ? alpha : beta;
    }
    _searchDeepSpace(playerId, cur) {
        const { state, _mainPlayerId } = this;
        if (cur >= this.MAX_DEPTH_DEEP)
            return this._score(playerId);
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (state.isDraw())
            return Number.MAX_VALUE;
        const candidate = state.topCandidate(playerId);
        if (candidate.score < this._CANDIDATE_SCORE_DEEP_MIN) {
            return state.score(playerId ^ 1, _mainPlayerId);
        }
        this._forward(candidate.posId, playerId);
        const result = this._searchDeepSpace(playerId ^ 1, cur + 1);
        this._revert(candidate.posId);
        return result;
    }
    _score(nextPlayerId) {
        const { state, _mainPlayerId } = this;
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (state.isDraw())
            return Number.MAX_VALUE;
        return state.score(nextPlayerId ^ 1, _mainPlayerId);
    }
    *_selectOneCandidate(candidates = [], size) {
        const firstCandidate = candidates[0];
        yield firstCandidate;
        const MAX_CANDIDATE_SCORE = firstCandidate.score;
        let temperature = this._CANDIDATE_SCORE_DEEP_MAX;
        const TEMPERATURE_DROP_RATE = 0.98;
        for (let i = 1; i < size; ++i) {
            const candidate = candidates[i];
            const delta = candidate.score - MAX_CANDIDATE_SCORE;
            const possibility = Math.exp(delta / temperature);
            if (Math.random() < possibility) {
                yield candidate;
            }
            temperature *= TEMPERATURE_DROP_RATE;
        }
    }
    _forward(posId, playerId) {
        this.context.forward(posId, playerId);
        this.state.forward(posId);
    }
    _revert(posId) {
        this.context.revert(posId);
        this.state.revert(posId);
    }
}

export { GomokuContext, GomokuCountMap, GomokuDirectionType, GomokuDirectionTypeBitset, GomokuDirectionTypes, GomokuDirections, GomokuSolution, GomokuState, GomokuStateCache, GomokuStateCompressor, createScoreMap };
